@@ -73,17 +73,11 @@ typedef struct
     uint4 maxsize;
 
     char output[MAXOUTPUTSIZE];
-
+    candidate_t *tmp_candidates;
+    boole utfaware;
 } textcat_t;
 
 
-typedef struct
-{
-    int score;
-    const char *name;
-} candidate_t;
-
-
 static int cmpcandidates(const void *a, const void *b)
 {
     const candidate_t *x = (const candidate_t *)a;
@@ -101,12 +95,36 @@ extern void textcat_Done(void *handle)
     {
         fp_Done(h->fprint[i]);
     }
+    if (h->tmp_candidates != NULL)
+    {
+        textcat_ReleaseClassifyFullOutput(h, h->tmp_candidates);
+    }
     free(h->fprint);
     free(h->fprint_disable);
     free(h);
 
 }
 
+extern int textcat_SetProperty(void *handle, textcat_Property property,
+                               sint4 value)
+{
+    textcat_t *h = (textcat_t *) handle;
+    switch (property)
+    {
+    case TCPROP_UTF8AWARE:
+        if ((value == TC_TRUE) || (value == TC_FALSE))
+        {
+            h->utfaware = value;
+            return 0;
+        }
+        return -2;
+        break;
+    default:
+        break;
+    }
+    return -1;
+}
+
 /** Replaces older function */
 extern void *textcat_Init(const char *conffile)
 {
@@ -120,6 +138,9 @@ extern void *textcat_Init(const char *conffile)
 extern void *special_textcat_Init(const char *conffile, const char *prefix)
 {
     textcat_t *h;
+    char *finger_print_file_name;
+    size_t finger_print_file_name_size;
+    size_t prefix_size;
     char line[1024];
     FILE *fp;
 
@@ -136,13 +157,23 @@ extern void *special_textcat_Init(const char *conffile, const char *prefix)
     h->size = 0;
     h->maxsize = 16;
     h->fprint = (void **)malloc(sizeof(void *) * h->maxsize);
-    h->fprint_disable = (unsigned char *)malloc(sizeof(unsigned char) * h->maxsize);
+    h->fprint_disable =
+        (unsigned char *)malloc(sizeof(unsigned char) * h->maxsize);
     /* added to store the state of languages */
+    h->tmp_candidates = NULL;
+    h->utfaware = TC_TRUE;
+
+    prefix_size = strlen(prefix);
+    finger_print_file_name_size = prefix_size + 1;
+    finger_print_file_name =
+        (char *)malloc(sizeof(char) * (finger_print_file_name_size + 1024));
+    finger_print_file_name[0] = '\0';
+    strcat(finger_print_file_name, prefix);
+
     while (wg_getline(line, 1024, fp))
     {
         char *p;
         char *segment[4];
-        char finger_print_file_name[512 + 1];
         int res;
 
         /*** Skip comments ***/
@@ -163,8 +194,7 @@ extern void *special_textcat_Init(const char *conffile, const char *prefix)
                 (void **)realloc(h->fprint, sizeof(void *) * h->maxsize);
             h->fprint_disable =
                 (unsigned char *)realloc(h->fprint_disable,
-                                            sizeof(unsigned char) *
-                                            h->maxsize);
+                                         sizeof(unsigned char) * h->maxsize);
         }
 
         /*** Load data ***/
@@ -173,52 +203,123 @@ extern void *special_textcat_Init(const char *conffile, const char *prefix)
             goto BAILOUT;
         }
 
-        /*** Check filename overflow ***/
-        finger_print_file_name[0] = finger_print_file_name[512] = '\0';
-        strncat(finger_print_file_name, prefix, 512);
-        strncat(finger_print_file_name, segment[0], 512);
-        if (finger_print_file_name[512] != '\0')
+        while (prefix_size + strlen(segment[0]) > finger_print_file_name_size)
         {
-            goto BAILOUT;
+            char *tmp;
+            size_t tmp_size = finger_print_file_name_size * 2;
+            tmp =
+                (char *)realloc(finger_print_file_name,
+                                sizeof(char) * (tmp_size + 1));
+            if (tmp == NULL)
+            {
+                free(finger_print_file_name);
+                finger_print_file_name_size = 0;
+                goto BAILOUT;
+            }
+            else
+            {
+                finger_print_file_name = tmp;
+                finger_print_file_name_size = tmp_size;
+            }
         }
+        finger_print_file_name[prefix_size] = '\0';
+        strcat(finger_print_file_name, segment[0]);
 
         if (fp_Read(h->fprint[h->size], finger_print_file_name, 400) == 0)
         {
+            textcat_Done(h);
             goto BAILOUT;
         }
         h->fprint_disable[h->size] = 0xF0;  /* 0xF0 is the code for enabled
-                                               languages, 0x0F is for disabled */
+                                               languages, 0x0F is for disabled 
+                                             */
         h->size++;
     }
 
+    free(finger_print_file_name);
+
     fclose(fp);
     return h;
 
   BAILOUT:
-    textcat_Done(h);            /* don't leak h */
     fclose(fp);
     return NULL;
 
 }
 
+extern candidate_t *textcat_GetClassifyFullOutput(void *handle)
+{
+    textcat_t *h = (textcat_t *) handle;
+    return (candidate_t *) malloc(sizeof(candidate_t) * h->size);
+}
+
+extern void textcat_ReleaseClassifyFullOutput(void *handle,
+                                              candidate_t * candidates)
+{
+    if (candidates != NULL)
+    {
+        free(candidates);
+    }
+}
+
 extern char *textcat_Classify(void *handle, const char *buffer, size_t size)
 {
     textcat_t *h = (textcat_t *) handle;
-    int minscore = MAXSCORE;
-    int threshold = minscore;
     char *result = h->output;
-    void *unknown;
     uint4 i, cnt;
 
-    candidate_t *candidates =
-        (candidate_t *) malloc(sizeof(candidate_t) * h->size);
+    if (h->tmp_candidates == NULL)
+    {
+        h->tmp_candidates = textcat_GetClassifyFullOutput(h);
+    }
+
+    cnt = textcat_ClassifyFull(h, buffer, size, h->tmp_candidates);
+
+    switch (cnt)
+    {
+    case TEXTCAT_RESULT_UNKOWN:
+        result = _TEXTCAT_RESULT_UNKOWN;
+        break;
+    case TEXTCAT_RESULT_SHORT:
+        result = _TEXTCAT_RESULT_SHORT;
+        break;
+    default:
+        {
+            const char *plimit = result + MAXOUTPUTSIZE;
+            char *p = result;
+
+            *p = '\0';
+            for (i = 0; i < cnt; i++)
+            {
+                p = wg_strgmov(p, "[", plimit);
+                p = wg_strgmov(p, h->tmp_candidates[i].name, plimit);
+                p = wg_strgmov(p, "]", plimit);
+            }
+        }
+    }
+
+    return result;
+}
+
+
+extern int textcat_ClassifyFull(void *handle, const char *buffer, size_t size,
+                                candidate_t * candidates)
+{
+    textcat_t *h = (textcat_t *) handle;
+    uint4 i, cnt = 0;
+    int minscore = MAXSCORE;
+    int threshold = minscore;
+
+    void *unknown;
 
     unknown = fp_Init(NULL);
+    fp_SetProperty(unknown, TCPROP_UTF8AWARE, h->utfaware);
+    fp_SetProperty(unknown, TCPROP_MINIMUM_DOCUMENT_SIZE, MINDOCSIZE);
     if (fp_Create(unknown, buffer, size, MAXNGRAMS) == 0)
     {
         /*** Too little information ***/
-        result = _TEXTCAT_RESULT_SHORT;
-        goto READY;
+        fp_Done(unknown);
+        return TEXTCAT_RESULT_SHORT;
     }
 
     /*** Calculate the score for each category. ***/
@@ -258,30 +359,17 @@ extern char *textcat_Classify(void *handle, const char *buffer, size_t size)
         }
     }
 
+    fp_Done(unknown);
     /*** The verdict ***/
     if (cnt == MAXCANDIDATES + 1)
     {
-        result = _TEXTCAT_RESULT_UNKOWN;
+        return TEXTCAT_RESULT_UNKOWN;
     }
     else
     {
-        const char *plimit = result + MAXOUTPUTSIZE;
-        char *p = result;
-
         qsort(candidates, cnt, sizeof(candidate_t), cmpcandidates);
-
-        *p = '\0';
-        for (i = 0; i < cnt; i++)
-        {
-            p = wg_strgmov(p, "[", plimit);
-            p = wg_strgmov(p, candidates[i].name, plimit);
-            p = wg_strgmov(p, "]", plimit);
-        }
+        return cnt;
     }
-  READY:
-    fp_Done(unknown);
-    free(candidates);
-    return result;
 }
 
 extern const char *textcat_Version(void)
